from collections import defaultdict
import neal
from .utils import simulator as simulator_utils
import itertools
from pyqubo import Array, Placeholder, Constraint, SubH
from snakes.nets import *
import snakes.plugins
snakes.plugins.load("gv", "snakes.nets", "nets")

def main(petrinet, feed_dict={'A': 1.0, 'B': 1.0, 'C': 1.0}):
    # NOTE: If you put 0 in a feed_dict, it won't be able to judge correctly.
    if not (feed_dict['A'] and feed_dict['B'] and feed_dict['C']):
        print('incorrent feed_dict')
        exit(1)

    solution_dict = {}
    start_time, end_time, feasible_solution, subH, bs_log = binary_search(petrinet, feed_dict)

    if feasible_solution:
        solution_dict = simulator_utils.format_solution_to_2Ddict(feasible_solution.sample)

    sorted_solution = sorted(solution_dict.items(), key=lambda x: x[0])
    return {'solution': {'start_time': start_time, 'end_time': end_time, 'energy': sum(subH.values()), 'spins': sorted_solution,
            'subH': { 'H_prece': subH['H_precedence'], 'H_conflict': subH['H_conflict'], 'H_firing': subH['H_firing']}},
            'topology': {'jobs': petrinet.jobs,
                         'processing_times': petrinet.pt,
                         'resources_by_trans': dict(petrinet.resources_trans_map),
                         'num_var': len(petrinet.preplaces_trans_map(resource=False)) * (sum(petrinet.pt.values()) - min(petrinet.pt.values()))
                         },
            'bs_log': bs_log,
            'mode': 'simulator'
            }


def binary_search(n, feed_dict):
    t_num = len(n.preplaces_trans_map(resource=False))
    a = Placeholder("A")
    b = Placeholder("B")
    c = Placeholder("C")
    machine_num = len(n.trans_resource_map)
    # Divide the sum of all task lengths by the number of fastest machines.
    earliest_time = sum(n.pt.values()) // machine_num
    # (The sum of all task lengths executed on the fastest machine) - (the smallest pt)
    latest_time = sum(n.pt.values()) - min(n.pt.values())
    feasible_solution = []
    feasible_subH = defaultdict(int)
    bs_cnt = 0
    bs_log = {}
    bs_log.setdefault('total_cnt', 0)
    bs_log.setdefault('cnt', {})

    while earliest_time <= latest_time:
        mid_time = (earliest_time + latest_time) // 2
        x = Array.create('x', (t_num, mid_time), 'BINARY')
        H_precedence = set_precedence(n, x, mid_time)
        H_conflict = set_conflict(n, x, mid_time)
        H_firing = set_firing(t_num, x, mid_time)
        H = SubH(a * H_precedence, 'H_precedence') + SubH(b * H_conflict, 'H_conflict') + SubH(c * H_firing, 'H_firing')
        model = H.compile()
        bqm = model.to_bqm(feed_dict=feed_dict)
        sampler = neal.SimulatedAnnealingSampler()
        sampleset = sampler.sample(bqm)
        decoded_samples = model.decode_sampleset(sampleset, feed_dict=feed_dict)
        decoded_solution = min(decoded_samples, key=lambda x: x.energy)
        broken = False if decoded_solution.energy == 0 else True

        if not broken:
            latest_time = mid_time - 1
            feasible_solution = decoded_solution
        else:
            earliest_time = mid_time + 1

        bs_log['cnt'].setdefault(bs_cnt, {'using_var_num': t_num * mid_time})
        bs_cnt += 1

    bs_log['total_cnt'] = bs_cnt

    # No existng feasible solution
    if not feasible_solution:
        return (-1, -1, [], {'H_precedence': decoded_solution.subh['H_precedence'],
                             'H_conflict': decoded_solution.subh['H_conflict'],
                             'H_firing': decoded_solution.subh['H_firing']},
                             bs_log
                )

    feasible_subH = {'H_precedence': feasible_solution.subh['H_precedence'],
                     'H_conflict': feasible_solution.subh['H_conflict'],
                     'H_firing': feasible_solution.subh['H_firing']}
    start_time, end_time = simulator_utils.calculate_processing_time(earliest_time, feasible_solution.sample, n.pt)

    return start_time, end_time, feasible_solution, feasible_subH, bs_log


def set_firing(t_num, x, time):
    H_firing = 0.0
    for i in range(t_num):
        H_firing += Constraint((1-sum(x[i, j] for j in range(time)))
                               ** 2, label=f"one_fired_t{i}", condition=lambda x: x == 0.0)
    return H_firing


def set_precedence(n, x, time):
    H_precedence = 0.0
    for j in range(time):
        for trans in n.preplaces_trans_map(resource=False):
            # Putting precedence constraints based on pre transition and post transition
            pre_t = int(re.sub("\\D", "", trans))

            if not(trans in n.postplaces_trans_map(resource=False)):
                continue

            post_p = n.postplaces_trans_map(resource=False)[trans]
            if not(post_p in n.posttrans_place_map):
                continue

            post_t = int(re.sub("\\D", "", n.posttrans_place_map[post_p]))
            for nj in range(time):
                if (j + n.pt[f"t{pre_t}"] > nj):
                    H_precedence += Constraint(x[pre_t, j] * x[post_t, nj],
                                               label=f"adjacent(t{pre_t},t{post_t})", condition=lambda x: x == 0.0)
    return H_precedence


def set_conflict(n, x, time):
    H_conflict = 0.0
    H_conflict_dict = defaultdict(int)
    pair = []
    for value in n.trans_resource_map.values():
        value = list(value)
        pair.append(list(itertools.combinations(value, 2)))

    for i, k in list(itertools.chain.from_iterable(pair)):
        i = int(re.sub("\\D", "", i))
        k = int(re.sub("\\D", "", k))

        for t in range(time):
            for nt in range(t, time):
                if (t + n.pt[f"t{i}"] > nt and n.pt[f"t{k}"] > 0):
                    H_conflict += Constraint(x[i, t] * x[k, nt],
                                             label=f"one_job_t{i}_and_t{k}")
                    H_conflict_dict[f"{k}{nt}{i}{t}"] = 1

                if ((t + n.pt[f"t{k}"] > nt) and ((f"{k}{t}{i}{nt}" in H_conflict_dict) == False and n.pt[f"t{i}"] > 0)):
                    H_conflict += Constraint(x[k, t] * x[i, nt],
                                             label=f"one_job_t{k}_and_t{i}", condition=lambda x: x == 0.0)
    return H_conflict



from pxi4jss.solver import main as solver
import snakes_utils

def main(snakes_net):
    # TODO: separate snakes_utils from here
    topology = snakes_utils.JSS(snakes_net, rflag='m')
    return solver(topology)






import re
import pandas as pd

def extract_variable(s):
    literal = re.split('[\[\]]', s)
    while '' in literal:
        literal.remove('')
    return literal

def calculate_processing_time(low_time, decoded_solution_sample, pt):
    if not decoded_solution_sample:
        return -1, -1

    start_time = low_time - 1
    solution_dict = format_solution_to_2Ddict(decoded_solution_sample)
    solution_df = pd.DataFrame(solution_dict).T

    # If the time size of solution_df and start_time do not match
    solution_df_last_column = len(solution_df.columns)-1
    if (solution_df[solution_df_last_column] == 0).all():
        for ci in range(solution_df_last_column, -1, -1):
            # check if a column is all zeroes
            if (solution_df[ci] == 0).all():
                solution_df.drop(columns=ci, inplace=True)
                start_time -= 1
            else:
                break

    max_start_time_transitions = solution_df[solution_df[start_time]
                                                == 1].index.values
    end_time = 0
    for t in max_start_time_transitions:
        end_time = max(start_time + (pt['t{}'.format(t)]), end_time)

    return start_time, end_time

def format_solution_to_2Ddict(solution):
    if not solution:
        return {}

    formated_solution = {}
    for k, v in solution.items():
        literal = extract_variable(k)
        time = int(literal[2])
        trans = int(literal[1])

        if not (trans in formated_solution):
            formated_solution[trans] = {}
        formated_solution[trans].update({time: v})

    # {0: {0: 1, 1: 0, 2: 0, 3: 0, 4: 0},
    #  1: {0: 0, 1: 0, 2: 1, 3: 0, 4: 0},
    #  2: {0: 0, 1: 0, 2: 0, 3: 0, 4: 1},
    #  3: {0: 1, 1: 0, 2: 0, 3: 0, 4: 0},
    #  4: {0: 0, 1: 0, 2: 0, 3: 1, 4: 0},
    #  5: {0: 0, 1: 0, 2: 0, 3: 0, 4: 1}}
    return formated_solution



import unittest
from pyqubo import Array, Placeholder
from snakes.nets import *
from pxi4jss import main as pxi4jss

class TestMain(unittest.TestCase):
    def setUp(self):
        snakes_net = self.init_petrinet()
        self.res = pxi4jss(snakes_net)

    def init_petrinet(self):
        n = PetriNet('Job Shop Scheduling')
        n.add_place(Place('p0', [1]))
        n.add_place(Place('p1'))
        n.add_place(Place('p2'))
        n.add_place(Place('p3'))
        n.add_place(Place('p4'))
        n.add_place(Place('p5', [1]))
        n.add_place(Place('p6'))
        n.add_place(Place('p7'))
        n.add_place(Place('p8'))
        n.add_place(Place('p9'))
        n.add_place(Place('p10', [1]))
        n.add_place(Place('p11'))
        n.add_place(Place('p12'))
        n.add_place(Place('p13'))
        n.add_place(Place('p14'))

        n.add_place(Place('m0'))
        n.add_place(Place('m1'))
        n.add_place(Place('m2'))

        pt = {'t1': '1',
              't2': '1',
              't3': '2',
              't0': '2',
              't4': '1',
              't6': '2',
              't7': '1',
              't5': '3',
              't8': '2',
              't9': '3',
              't10': '2',
              't11': '2'}

        t_num = 12  # Number of operations
        for i in range(t_num):
            t_name = f't{i}'
            t_pt = pt[t_name]
            n.add_transition(Transition(t_name, Expression(t_pt)))

        start = 0
        end = per_p = 4
        job_num = 3
        for j in range(job_num):
            for i in range(start, end):
                n.add_input('p{}'.format(i+j), 't{}'.format(i), Variable('x'))
                n.add_output('p{}'.format(i+j+1),
                             't{}'.format(i), Variable('x'))

            start = end
            end += per_p

        # Machines
        m_dict = {'m1': {'t2', 't5', 't6', 't1'},
                  'm2': {'t7', 't3', 't11', 't10'},
                  'm0': {'t8', 't9', 't4', 't0'}}
        for m, t in m_dict.items():
            for e in t:
                n.add_input(m, e, Variable('x'))
                n.add_output(m, e, Variable('x'))

        return n

    def test_starttime_is_greater_than(self):
        if self.res['solution']['energy'] == 0:
            self.assertTrue(self.res['solution']['start_time'] >= 11 )

    def test_endtime_is_greater_than(self):
        if self.res['solution']['energy'] == 0:
            self.assertTrue(self.res['solution']['end_time'] >= 13 )

    def test_calc_energy_of_subH(self):
        if self.res['solution']['energy'] == 0:
            subH = self.res['solution']['subH']
            self.assertEqual(subH['H_conflict'], 0.0)
            self.assertEqual(subH['H_prece'], 0.0)
            self.assertEqual(subH['H_firing'], 0.0)

    def test_extract_pt(self):
        if self.res['solution']['energy'] == 0:
            expected = {'t1': 1,
                        't2': 1,
                        't3': 2,
                        't0': 2,
                        't4': 1,
                        't6': 2,
                        't7': 1,
                        't5': 3,
                        't8': 2,
                        't9': 3,
                        't10': 2,
                        't11': 2}

            self.assertEqual(self.res['topology']['processing_times'], expected)

    def test_extract_resources_by_trans(self):
        if self.res['solution']['energy'] == 0:
            expected = {'t0': {'m0'}, 't1': {'m1'}, 't2': {'m1'}, 't3': {'m2'}, 't4': {'m0'}, 't5': {
                'm1'}, 't6': {'m1'}, 't7': {'m2'}, 't8': {'m0'}, 't9': {'m0'}, 't10': {'m2'}, 't11': {'m2'}}

            self.assertEqual(self.res['topology']['resources_by_trans'], expected)

    def test_calc_num_of_var(self):
        expected = 252
        self.assertEqual(self.res['topology']['num_var'], expected)

    def test_extract_jobs(self):
        expected = {0: ['t0', 't1', 't2', 't3'],
                    1: ['t4', 't5', 't6', 't7'],
                    2: ['t8', 't9', 't10', 't11']}

        actual = self.res['topology']['jobs']
        self.assertTrue(self.assert_equal_jobs(expected, actual))

    # Without considering the order of the keys
    def assert_equal_jobs(self, jobs, jobs2):
        def subtract_list(lst1, lst2):
            lst = lst1.copy()
            for element in lst2:
                try:
                    lst.remove(element)
                except ValueError:
                    return False
            if not lst:
                return True
            return False

        jobs_values = list(jobs.values())
        jobs2_values = list(jobs2.values())
        if not (subtract_list(jobs_values, jobs2_values) or subtract_list(jobs2_values, jobs_values)):
            return False

        return True

if __name__ == "__main__":
    unittest.main()



from setuptools import setup, find_packages
from os import path

here = path.abspath(path.dirname(__file__))
with open(path.join(here, 'README.md'), 'r', encoding='utf-8') as f:
    readme = f.read()

setup(
    name="pxi4jss",
    packages=find_packages(),
    install_requires=['snakes-utils>=2.0.2', 'dwave-neal', 'pandas', 'pyqubo'],
    version="1.0",
    license="MIT",
    author='kanekou',
    author_email='k208580@ie.u-ryukyu.ac.jp',
    url='https://github.com/kanekou/pxi4jss',
    description='It is modeling tool for automatically generating Ising models from JSS Petri nets.',
    long_description=readme,
    long_description_content_type="text/markdown",
    classifiers=[
        'Programming Language :: Python :: 3.7',
        'License :: OSI Approved :: MIT License',
    ],
)
